class PointerString:
    _string: str
    _pointer: int

    @property
    def pointer(self) -> int:
        return self._pointer
    
    @pointer.setter
    def pointer(self, value: int) -> None:
        self._check_if_value_is_correct_pointer_and_raise_error_if_not(value)
        self._pointer = value

    @property
    def string(self) -> str:
        return self._string

    def __init__(self, string):
        self._string = string
        self.pointer = 0

    def _check_if_value_is_correct_pointer_and_raise_error_if_not(self, value):
        if not (0 <= value < len(self._string)):
            raise IndexError("pointer out of the string")

    def move(self, steps: int) -> str:
        new: int = self.pointer + steps
        self.pointer = new
        return self.get_symb()
    
    def get_symb(self, offset: int = 0) -> str:
        index: int = self.pointer+offset
        self._check_if_value_is_correct_pointer_and_raise_error_if_not(index)
        return self._string[index]

    def __len__(self): 
        return len(self._string)
    
    def __copy__(self):
        new: PointerString = PointerString(self._string)
        new.pointer = self.pointer
        return new


from pathlib import Path
from typing import Union
from .abstract_template import AbstractTemplate

class FileTemplate(AbstractTemplate):
    def __init__(self, filename: Union[str, Path]):
        file = open(filename, "r")
        self._string = file.read()
        self.name = str(filename)
        super().__init__()


from typing import Callable, Optional, Union
from .file_template import FileTemplate
from pathlib import Path

class TemplatesLoader:
    templates: dict[str, FileTemplate]
    root_path = Path(".").absolute()

    def __init__(self):
        self.templates = {}

    def load_template(self, name: Union[str, Path]):
        self.templates[str(name)] = FileTemplate(name)

    def unload_template(self, name: str):
        del self.templates[name]

    def recursively_load_folder(self, path_str: str, pattern: str = "*", filter: Optional[Callable[[Path], bool]] = None):
        loaded: list[Path] = []
        for path in Path(path_str).rglob(pattern):
            try:
                path = path.absolute().relative_to(self.root_path)
            except ValueError as e:
                raise ValueError("Directory should be the child directory of the current working directory") from e
            if not filter or filter(path):
                self.load_template(path)
                loaded.append(path)
        return loaded
    
    def get_template(self, name: str):
        if name not in self.templates:
            self.load_template(name)
        return self.templates[name]
    







from abc import abstractclassmethod, abstractmethod, abstractstaticmethod
from dataclasses import dataclass
from typing import Iterable, Type
from .pointer_string import PointerString

class InnerPrerendererError(Exception):
    def __init__(self):
        super().__init__("inner prerenderer error")

@dataclass
class PrerenderZonesParametersBase: 
    pointer_original: PointerString


class AbstractPrerenderZone:
    @abstractclassmethod
    def entry_condition(cls, parameters: PrerenderZonesParametersBase) -> bool:
        """If the parameters.pointer_original.get_symb(0) should be the first symbol of the zone, returns True, else returns False.
        Doesn't move the pointer in the parameters.pointer_original PointerString"""
        raise NotImplementedError()
    
    @abstractmethod
    def process(self, parameters: PrerenderZonesParametersBase) -> str:
        """Returns the result of the parsing of the zone that starts from parameters.pointer_original.get_symb(0) and moves the pointer to the last symbol of the zone"""
        raise NotImplementedError()



class AbstractPrerenderZoneWithSubzones(AbstractPrerenderZone):
    _possible_subzone_types: Iterable[Type[AbstractPrerenderZone]]
    
    @abstractclassmethod
    def _entry_condition(cls, parameters: PrerenderZonesParametersBase) -> bool:
        """If the parameters.pointer_original.get_symb(0) should be the first symbol of the zone, returns True, else returns False.
        Doesn't move the pointer in the parameters.pointer_original PointerString"""
        raise NotImplementedError()

    @abstractmethod
    def _entry(self, parameters: PrerenderZonesParametersBase) -> str:
        """Returns the result of the parsing of the start of the zone whose first symbol is parameters.pointer_string.get_symb(0)."""
        raise NotImplementedError()
    
    @abstractmethod
    def _process_symbol(self, parameters: PrerenderZonesParametersBase) -> str:
        """Returns the result of the parsing of the symbol parameters.pointer_string.get_symb(0) that's inside of the zone."""
        raise NotImplementedError()

    @abstractmethod
    def _exit_condition(self, parameters: PrerenderZonesParametersBase) -> bool:
        "Returns True if zone's last symbol is parameters.pointer_original.get_symb(0), else returns False"
        raise NotImplementedError()
    
    @abstractmethod
    def _exit(self, parameters: PrerenderZonesParametersBase) -> str:
        """Returns the result of the parsing of the end of the zone whose last symbol is parameters.pointer_string.get_symb(0)."""
        raise NotImplementedError()

    @abstractmethod
    def _subzone_absence_entry(self, parameters: PrerenderZonesParametersBase) -> str:
        """Returns the result of the parsing of the start of the subzone absence whose first symbol is parameters.pointer_string.get_symb(0)."""
        raise NotImplementedError()

    @abstractmethod
    def _subzone_absence_exit(self, parameters: PrerenderZonesParametersBase) -> str:
        """Returns the result of the parsing of the end of the subzone absence whose last symbol is parameters.pointer_string.get_symb(0)."""
        raise NotImplementedError()

    @abstractmethod
    def _process_index_error(self, e: IndexError, parameters: PrerenderZonesParametersBase) -> None:
        raise NotImplementedError()

    @classmethod
    def entry_condition(cls, parameters: PrerenderZonesParametersBase) -> bool:
        return cls._entry_condition(parameters)
    
    def process(self, parameters: PrerenderZonesParametersBase) -> str:
        result: str = ""
        result += self._entry(parameters)
        original: PointerString = parameters.pointer_original
        is_zone_absence_entry: bool = True
        while 1:
            Subzone: Type[AbstractPrerenderZone]
            for Subzone in self._possible_subzone_types:
                if Subzone.entry_condition(parameters):
                    if not is_zone_absence_entry:
                        result += self._subzone_absence_exit(parameters)
                    subzone = Subzone()
                    result += subzone.process(parameters)
                    is_zone_absence_entry = True
                    break
            else:
                if is_zone_absence_entry:
                    result += self._subzone_absence_entry(parameters)
                result += self._process_symbol(parameters)
                is_zone_absence_entry = False
            try:
                original.move(1)
            except IndexError as e:
                self._process_index_error(e, parameters)
                raise InnerPrerendererError()
            if self._exit_condition(parameters):
                result += self._process_symbol(parameters)
                if not is_zone_absence_entry:
                    result += self._subzone_absence_exit(parameters)
                result += self._exit(parameters)
                break #TODO: move exit condition checking to the start of the while loop to fix the bag of one symbol template
        return result


from types import CodeType
from typing import Any, Iterable, Type
from .abstract_prerender_zone import AbstractPrerenderZone, AbstractPrerenderZoneWithSubzones, PrerenderZonesParametersBase
from .pointer_string import PointerString

TAB = " "

class PrerenderZonesParameters(PrerenderZonesParametersBase):
    tabs_number: int
    template_name: str

    def __init__(self, string: str, template_name: str):
        self.tabs_number = 0
        # try:
        #     self.pointer_original = PointerString(string)
        # except IndexError:
        #     raise TemplateSyntaxTemplateIsEmptyError(template_name) TODO: uncomment to fix an empty template bug
        self.template_name = template_name

class InnerTemplaterError(Exception):
    def __init__(self):
        super().__init__("inner templater error")

class TemplateSyntaxError(SyntaxError):
    def __init__(self, name: str):
        self._template_name = name
    
    def set_message(self, message: str):
        super().__init__(f"in the \"{self._template_name}\" template: " + message)

class TemplateSyntaxCurlyBracketsWereNotClosedError(TemplateSyntaxError):
    def __init__(self, name):
        super().__init__(name)
        self.set_message(f"curly brackets were not closed")

class TemplateSyntaxTabsError(TemplateSyntaxError):
    def __init__(self, name):
        super().__init__(name)
        self.set_message("\"{{-}}\" should be used in pairs with {{+}} above")

class TemplateSyntaxNewLineInCodeBlockError(TemplateSyntaxError):
    def __init__(self, name):
        super().__init__(name)
        self.set_message("new lines in a code zone are not allowed. Use several code zones instead.")

class TemplateSyntaxCompilationError(TemplateSyntaxError):
    def __init__(self, name):
        super().__init__(name)
        self.set_message("compilation error")

# class TemplateSyntaxTemplateIsEmptyError(TemplateSyntaxError):
#     def __init__(self, name):
#         super().__init__(name)
#         self.set_message("emply templates are not allowed") TODO: uncomment to fix an empty template bug


class PrerenderRootZone(AbstractPrerenderZoneWithSubzones):
    _possible_subzone_types: list[Type[AbstractPrerenderZone]] = []
    symbols_to_protect: Iterable[str] = ["{", "\\"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._possible_subzone_types.append(PrerenderChangeTabsNumber)
        self._possible_subzone_types.append(PrerenderCodeZone)

    @classmethod
    def _entry_condition(cls, parameters: PrerenderZonesParameters) -> bool:
        return True

    def _entry(self, parameters: PrerenderZonesParameters) -> str:
        return ""

    def _process_symbol(self, parameters: PrerenderZonesParameters) -> str:
        result: str = parameters.pointer_original.get_symb(0)
        try:
            next_symb = parameters.pointer_original.get_symb(1)
        except IndexError:
            pass
        else:
            if result == "\\" and next_symb in self.symbols_to_protect:
                result += parameters.pointer_original.move(1)
        return result

    def _exit_condition(self, parameters: PrerenderZonesParameters) -> bool:
        return parameters.pointer_original.pointer == len(parameters.pointer_original.string)-1

    def _exit(self, parameters: PrerenderZonesParameters) -> str:
        return ""

    def _subzone_absence_entry(self, parameters: PrerenderZonesParameters) -> str:
        return parameters.tabs_number*TAB+"insert(\"\"\""

    def _subzone_absence_exit(self, parameters: PrerenderZonesParameters) -> str:
        return "\"\"\")\n"

    def _process_index_error(self, e: IndexError, parameters: PrerenderZonesParameters) -> None:
        raise InnerTemplaterError() from e
    

class PrerenderChangeTabsNumber(AbstractPrerenderZone):
    @classmethod
    def entry_condition(cls, parameters: PrerenderZonesParameters) -> bool:
        try:
            if parameters.pointer_original.get_symb(0) == parameters.pointer_original.get_symb(1) == "{":
                if parameters.pointer_original.get_symb(3) == parameters.pointer_original.get_symb(4) == "}":
                    return True
        except IndexError:
            pass
        return False

    def process(self, parameters: PrerenderZonesParameters) -> str:
        symb2 = parameters.pointer_original.get_symb(2)
        if symb2 == "+":
            parameters.tabs_number+=1
        elif symb2 == "-":
            parameters.tabs_number-=1
        parameters.pointer_original.move(4)
        return ""


class PrerenderCodeZone(AbstractPrerenderZoneWithSubzones):
    _possible_subzone_types: Iterable[Type[AbstractPrerenderZone]] = []
    exit: bool = False
    symbols_to_protect = ["}", "\\"]
    @classmethod
    def _entry_condition(cls, parameters: PrerenderZonesParameters) -> bool:
        try:
            return parameters.pointer_original.get_symb(0) == parameters.pointer_original.get_symb(1) == "{"
        except IndexError:
            return False

    def _entry(self, parameters: PrerenderZonesParameters) -> str:
        try:
            parameters.pointer_original.move(2)
        except IndexError:
            raise TemplateSyntaxCurlyBracketsWereNotClosedError(parameters.template_name)
        return parameters.tabs_number*TAB
    
    def _process_symbol(self, parameters: PrerenderZonesParameters) -> str:
        symb: str = parameters.pointer_original.get_symb(0)
        result = symb
        if symb == "\n":
            raise TemplateSyntaxNewLineInCodeBlockError(parameters.template_name)
                
        elif self.exit:
            return ""
        
        try:
            if symb == "\\" and parameters.pointer_original.get_symb(1) in self.symbols_to_protect:
                result += parameters.pointer_original.move(1)
        except IndexError:
            raise TemplateSyntaxCurlyBracketsWereNotClosedError(parameters.template_name)

        return result
    
    def _exit_condition(self, parameters: PrerenderZonesParameters) -> bool:
        try:
            if parameters.pointer_original.get_symb(0) == parameters.pointer_original.get_symb(1) == "}":
                self.exit = True
                return True
        except IndexError:
            raise TemplateSyntaxCurlyBracketsWereNotClosedError(parameters.template_name)
        return False

    def _exit(self, parameters: PrerenderZonesParameters) -> str:
        parameters.pointer_original.move(1)
        return "\n"

    def _subzone_absence_entry(self, parameters: PrerenderZonesParameters) -> str:
        return ""
    
    def _subzone_absence_exit(self, parameters: PrerenderZonesParameters) -> str:
        return ""
    
    def _process_index_error(self, e: IndexError, parameters: PrerenderZonesParameters) -> None:
        raise TemplateSyntaxCurlyBracketsWereNotClosedError(parameters.template_name)

class AbstractTemplate:
    _string: str
    name: str
    _code: CodeType

    def __init__(self):
        self._prerender()
    
    def _exec_code(self, globals: dict) -> None:
        exec(self._code, globals)

    def _prerender(self) -> None:
        root: PrerenderRootZone = PrerenderRootZone()
        parameters: PrerenderZonesParameters = PrerenderZonesParameters(self._string, self.name)
        source_code = root.process(parameters)
        try:
            self._code = compile(source_code, self.name, "exec")
        except Exception as e:
            raise TemplateSyntaxCompilationError(parameters.template_name) from e

    def render(self, context: Any) -> str:
        result: str = ""

        def insert(string: Any) -> None:
            nonlocal result
            result += str(string)

        self._exec_code({
            "context": context, 
            "__name__": self.name, 
            "__doc__": None,
            "insert": insert,
            })
        return result